"""Contains Query builder class"""

from typing import Any, Dict, List, Optional, Tuple, Union

from .enums import Order
from .filters import Filter, TimeDimension
from .objects import Dimension, Measure, Segment


class Query:
    """Class to build Cube.js queries"""

    def __init__(
        self,
        measures: List[Measure],
        time_dimensions: List[TimeDimension],
        dimensions: Optional[List[Dimension]] = None,
        filters: Optional[List[Filter]] = None,
        segments: Optional[List[Segment]] = None,
        limit: int = 10_000,
        offset: int = 0,
        order: Optional[List[Tuple[Union[Dimension, Measure], Order]]] = None,
        timezone: str = "UTC",
        ungrouped: bool = False,
    ) -> None:
        """Initializer

        Args:
            measures: List of measures
            time_dimensions: List of time dimensions for filtering/grouping
            dimensions: List of dimensions for grouping
            filters: List of filters
            segments: List of segments
            limit: Limit the result set
            offset: Offset to control pagination
            order: Ordering configuration for the result set
            timezone: Local timezone to reflect in the data
            ungrouped: Whether to not GROUP BY in the query

        """
        self.measures = measures
        self.time_dimensions = time_dimensions
        self.dimensions = dimensions
        self.filters = filters
        self.segments = segments
        self.limit = limit
        self.offset = offset
        self.order = order
        self.timezone = timezone
        self.ungrouped = ungrouped

    def serialize(self) -> Dict[str, Any]:
        """Serialize the object for building a query

        Returns:
            query dict

        """
        result = {
            "measures": [m.serialize() for m in self.measures],
            "timeDimensions": [d.serialize() for d in self.time_dimensions],
            "limit": self.limit,
            "offset": self.offset,
            "timezone": self.timezone,
            "ungrouped": self.ungrouped,
        }
        if self.dimensions:
            result["dimensions"] = [d.serialize() for d in self.dimensions]
        if self.filters:
            result["filters"] = [f.serialize() for f in self.filters]
        if self.segments:
            result["segments"] = [s.serialize() for s in self.segments]
        if self.order:
            result["order"] = [
                (obj.serialize(), order.value) for obj, order in self.order
            ]

        return result



"""Contains classes for working with filters"""

from typing import Any, Dict, List, Optional, Union

from .enums import FilterOperator, TimeGranularity
from .objects import Dimension, Measure

__ALL__ = ["DateRange", "TimeDimension", "Filter", "Or", "And"]


class DateRange:
    """Represents an absolute or relative date range for filtering"""

    def __init__(
        self,
        start_date: Optional[str] = None,
        end_date: Optional[str] = None,
        relative: Optional[str] = None,
    ) -> None:
        """Initializer

        Args:
            start_date: Start date for the absolute range. Format `YYYY-MM-DD`
                or `YYYY-MM-DDTHH:mm:ss.SSS`.
            end_date: End date for the absolute range. Format `YYYY-MM-DD`
                or `YYYY-MM-DDTHH:mm:ss.SSS`.
            relative: Relative date range, e.g. "this month"

        """
        assert (start_date is not None and end_date is not None) or (
            relative is not None
        )
        self.start_date = start_date
        self.end_date = end_date
        self.relative = relative

    def serialize(self) -> Union[List[str], str]:
        """Serialize the object for building a query

        Raises:
            ValueError: if the start and end date or relative date range was not provided

        Returns:
            list of (start date, end date) or the relative date range

        """
        if self.start_date is not None and self.end_date is not None:
            return [self.start_date, self.end_date]
        elif self.relative:
            return self.relative
        else:
            raise ValueError("Expected (start date, end date) or relative")


class TimeDimension:
    """Represents the combo of a time dimension and filter"""

    def __init__(
        self,
        dimension: Dimension,
        date_range: DateRange,
        compare_date_range: Optional[List[DateRange]] = None,
        granularity: TimeGranularity = TimeGranularity.null,
    ) -> None:
        """Initializer

        Args:
            dimension: Time dimension reference
            date_range: Absolute or relative date range
            compare_date_range: An array of date ranges to compare a measure change over
                previous period
            granularity: A granularity for a time dimension

        """
        self.dimension = dimension
        self.date_range = date_range
        self.compare_date_range = compare_date_range
        self.granularity = granularity

    def serialize(self) -> Dict[str, Any]:
        """Serialize the object for building a query

        Returns:
            dict of properties

        """
        result = {
            "dimension": self.dimension.serialize(),
            "dateRange": self.date_range.serialize(),
            "granularity": self.granularity.value,
        }
        if self.compare_date_range:
            result["compareDateRange"] = [
                dr.serialize() for dr in self.compare_date_range
            ]

        return result


class Filter:
    """Represents a data filter"""

    def __init__(
        self,
        member: Union[Dimension, Measure],
        operator: FilterOperator,
        values: Optional[List[Any]] = None,
    ) -> None:
        """Initializer

        Args:
            member: Dimension or measure to be used in the filter
                * When you filter on a dimension, you are restricting the raw data
                  before any calculations are made.
                * When you filter on a measure, you are restricting the results after
                  the measure has been calculated.
            operator: An operator to be used in the filter.
            values: An array of values for the filter.

        """
        self.member = member
        self.operator = operator
        self.values = values

    def serialize(self) -> Dict[str, Any]:
        """Serialize the object for building a query

        Returns:
            dict of properties

        """
        return {
            "member": self.member.serialize(),
            "operator": self.operator.value,
            "values": [str(value) for value in self.values],
        }


class BooleanExpression:
    """Represents a boolean operator in a filter"""

    operator: str

    def __init__(self, *operands: Union["BooleanExpression", Filter]) -> None:
        """Initializer

        Args:
            operands: An array with two or more filters or other boolean operators

        """
        self.operands = operands

    def serialize(self) -> Dict[str, List[Any]]:
        """Serialize the object for building a query

        Returns:
            dict of properties

        """
        return {self.operator: [o.serialize() for o in self.operands]}


class Or(BooleanExpression):
    """Boolean operator for `or`"""

    operator = "or"


class And(BooleanExpression):
    """Boolean operator for `and`"""

    operator = "and"



"""Contains objects used to build a Cube.js query"""

__ALL__ = ["Cube"]


class CubeObject:
    """Base class for objects defined in a cube.

    An instance of this object represents a *reference* to an object in a cube.js
    configuration.
    """

    def __init__(self, cube: "Cube", name: str) -> None:
        """Init

        Args:
            cube: Cube factory object
            name: Name of the cube object

        """
        self._cube = cube
        self.name = name

    def __eq__(self, o: object) -> bool:
        """Checks equality between CubeObject instances"""
        return (
            isinstance(o, self.__class__)
            and o._cube == self._cube
            and o.name == self.name
        )

    def __str__(self) -> str:
        """String representation of the CubeObject"""
        return f"{self._cube.name}.{self.name}"

    def serialize(self) -> str:
        """Serialize the object in order to send a query"""
        return str(self)


class Measure(CubeObject):
    """Measure reference"""

    pass


class Dimension(CubeObject):
    """Dimension reference"""

    pass


class Segment(CubeObject):
    """Segment reference"""

    pass


class Cube:
    """Factory that aliases object references with the correct cube name"""

    def __init__(self, name: str) -> None:
        """Initializer

        Args:
            name: Cube configuration name

        """
        self.name = name

    def __eq__(self, o: object) -> bool:
        """Checks equality between Cube objects"""
        return isinstance(o, self.__class__) and o.name == self.name

    def __repr__(self) -> str:
        """REPR representation"""
        return f'{self.__class__.__name__}(name="{self.name}")'

    def __str__(self) -> str:
        """String representation"""
        return self.name

    def measure(self, name: str) -> Measure:
        """Factory method for building a measure field reference

        Args:
            name: Measure name

        Returns:
            measure object reference

        """
        return Measure(self, name)

    def dimension(self, name: str) -> Dimension:
        """Factory method for building a dimension field reference

        Args:
            name: Dimension name

        Returns:
            dimension object reference

        """
        return Dimension(self, name)

    def segment(self, name: str) -> Segment:
        """Factory method for building a segment field reference

        Args:
            name: Segment name

        Returns:
            segment object reference

        """
        return Segment(self, name)



"""Contains the Cube.js API client"""

from datetime import datetime, timedelta
from typing import Any, Dict, Optional

import backoff
import httpx
import jwt

from .query import Query


class CubeClient:
    """Cube.js API client"""

    def __init__(
        self,
        host: str = "http://localhost:4000",
        base_path: str = "/cubejs-api",
        secret: Optional[str] = None,
        load_request_timeout: float = 30.0,
        token_ttl_hours: int = 1,
    ) -> None:
        """Initializer

        Args:
            host: Cube.js API host
            base_path: Cube.js API base path
            secret: Secret for signing tokens. Set to None to skip authentication.
            load_request_timeout: Timeout in seconds to wait for load responses
            token_ttl_hours: TTL in hours for the token lifetime

        """
        self._secret = secret
        self._load_request_timeout = load_request_timeout
        self._token_ttl_hours = token_ttl_hours
        self._http_client = httpx.AsyncClient(
            base_url=f"{host.rstrip('/')}/{base_path.strip('/')}"
        )

        self._token = None

    def _get_signed_token(self) -> Optional[str]:
        """Get or refresh the authentication token

        Returns:
            token or None if no secret was configured

        """
        if not self._secret:
            return None

        now = datetime.now()
        if not self._token or self._token_expiration <= now:
            self._token_expiration = now + timedelta(hours=self._token_ttl_hours)
            self._token = jwt.encode(
                {"exp": self._token_expiration}, self._secret, algorithm="HS256"
            )

        return self._token

    @property
    def token(self) -> Optional[str]:
        """Alias for getting the current token value"""
        return self._get_signed_token()

    async def load(self, query: Query) -> Dict[str, Any]:
        """Get the data for a query.

        Args:
            query: Query object

        Returns:
            dict with properties:
            * query -- The query passed via params
            * data -- Formatted dataset of query results
            * annotation -- Metadata for query. Contains descriptions for all query
                items.
                * title -- Human readable title from data schema.
                * shortTitle -- Short title for visualization usage (ex. chart overlay)
                * type -- Data type

        """
        return await self._request(
            "post",
            "/v1/load",
            body={"query": query.serialize()},
            timeout=self._load_request_timeout,
        )

    @backoff.on_exception(
        backoff.expo, httpx.RequestError, max_tries=8, jitter=backoff.random_jitter
    )
    async def _request(
        self, method: str, path: str, body: Optional[Any] = None, timeout: float = 5.0
    ):
        """Make API request to Cube.js server

        Args:
            method: HTTP method
            path: URL path
            body: Body to send with the request, if applicable
            timeout: Request timeout in seconds

        Returns:
            response data

        """
        headers = {}
        if self.token:
            headers["Authorization"] = self.token

        async with self._http_client as client:
            response = await client.request(
                method, path, json=body, headers=headers, timeout=timeout
            )
            response.raise_for_status()
            return response.json()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cubejsclientasync']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.1.0,<3.0.0', 'backoff>=1.10.0,<2.0.0', 'httpx>=0.18.2,<0.19.0']

setup_kwargs = {
    'name': 'cubejsclientasync',
    'version': '0.1.0',
    'description': 'Async Python Cube.js client',
    'long_description': '# cubejsclientasync\n\n[![](https://img.shields.io/pypi/v/cubejsclientasync.svg)](https://pypi.org/pypi/cubejsclientasync/) [![License](https://img.shields.io/badge/License-BSD%203--Clause-blue.svg)](https://opensource.org/licenses/BSD-3-Clause)\n\nAsync Python Cube.js client\n\nFeatures:\n\n- Cube.js API client that makes async requests\n- Rich objects for building queries with measures, dimensions, etc.\n\nTable of Contents:\n\n- [Installation](#installation)\n- [Development](#development)\n\n## Installation\n\ncubejsclientasync requires Python 3.6 or above.\n\n```bash\npip install cubejsclientasync\n```\n\n## Development\n\nTo develop cubejsclientasync, install dependencies and enable the pre-commit hook:\n\n```bash\npip install pre-commit poetry\npoetry install\npre-commit install\n```\n\nTo run tests:\n\n```bash\npoetry shell\npytest\n```\n',
    'author': 'Jonathan Drake',
    'author_email': 'jdrake@narrativescience.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NarrativeScience/cubejs-client-async',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.5,<4.0.0',
}


setup(**setup_kwargs)



"""Contains enums"""

import enum


@enum.unique
class TimeGranularity(enum.Enum):
    """Time granularity"""

    second = "second"
    minute = "minute"
    hour = "hour"
    day = "day"
    week = "week"
    month = "month"
    year = "year"
    null = None


@enum.unique
class Order(enum.Enum):
    """Result ordering directions"""

    asc = "asc"
    desc = "desc"


@enum.unique
class FilterOperator(enum.Enum):
    """Operators used in filters"""

    equals = "equals"
    not_equals = "notEquals"
    contains = "contains"
    not_contains = "notContains"
    gt = "gt"
    gte = "gte"
    lt = "lt"
    lte = "lte"
    is_set = "set"
    not_set = "notSet"
    in_date_range = "inDateRange"
    not_in_date_range = "notInDateRange"
    before_date = "beforeDate"
    after_date = "afterDate"
